unit ScanBitmaps;

interface

uses
  Graphics, TheseUtils;

type
  TDotStyle = (dtSolid, dtOffset, dtSpecular, dtFaded, dtDirectional, dtBidirectional);
  TDotStyles = set of TDotStyle;

  TDotParams = record
    Color: TColor;
    Style: TDotStyles;
    SolidColor: TColor;
    OffX, OffY: TReal;
    SpecularColor: TColor;
    FadeAlpha: TReal;
    DirX, DirY: TReal;
    Narrowness: TReal;
  end;
  PDotParams = ^TDotParams;

  TScanBitmap = class(TBitmap)
  protected
    ScanTable: array of PBGRFColor;

    function GetScanPixel(X, Y: Integer): TColor; virtual;
    procedure SetScanPixel(X, Y: Integer; const Value: TColor); virtual;

  public
    constructor Create; override;
    property ScanPixel[X, Y: Integer]: TColor read GetScanPixel write SetScanPixel;
    function FindScanPixel(X, Y: Integer): PBGRFColor; virtual;
    function FindScanPixelOrExcept(X, Y: Integer): PBGRFColor; virtual;
    procedure ClearTable;
    procedure SubRenderPixel(X, Y: TReal; Color: TColor);
    procedure DrawDot(CenterX, CenterY, Rad: TReal; Color: TColor; Style: TDotStyles = []; SolidColor: TColor = 0; OffX: TReal = 0; OffY: TReal = 0; SpecularColor: TColor = 0; FadeAlpha: TReal = 1; DirX: TReal = 0; DirY: TReal = -1; Narrowness: TReal = 0); overload;
    procedure DrawDot(CenterX, CenterY, Rad: TReal; const Params: TDotParams); overload;
  end;

implementation

uses
  SysUtils, Math;

{ TScanBitmap }

procedure TScanBitmap.ClearTable;
begin
  SetLength(ScanTable, 0);
end;

function TScanBitmap.FindScanPixel(X, Y: Integer): PBGRFColor;
begin
  if (X < 0)
  or (Y < 0)
  or (X >= Width)
  or (Y >= Height)
  then begin
    Result := nil;
    Exit;
  end;
  if Length(ScanTable) <> Height then begin
    SetLength(ScanTable, 0);
    SetLength(ScanTable, Height);
    FillChar(ScanTable[0], Length(ScanTable) * SizeOf(ScanTable[0]), 0);
  end;
  if ScanTable[Y] <> nil then begin
    Result := ScanTable[Y];
  end
  else begin
    Result := ScanLine[Y];
    ScanTable[Y] := Result;
  end;
  Inc(Result, X);
end;

function TScanBitmap.FindScanPixelOrExcept(X, Y: Integer): PBGRFColor;
begin
  Result := FindScanPixel(X, Y);
  if Result = nil then raise Exception.Create('Pixel (' + IntToStr(X) + ',' + IntToStr(Y) + ') is off the bitmap.');
end;

function TScanBitmap.GetScanPixel(X, Y: Integer): TColor;
begin
  Result := FindScanPixelOrExcept(X, Y).Color and $FFFFFF;
end;

procedure TScanBitmap.SetScanPixel(X, Y: Integer; const Value: TColor);
begin
  FindScanPixelOrExcept(X, Y).Color := Value;
end;



constructor TScanBitmap.Create;
begin
  inherited;

  PixelFormat := pf32bit;
end;

procedure TScanBitmap.DrawDot(CenterX, CenterY, Rad: TReal; Color: TColor; Style: TDotStyles = []; SolidColor: TColor = 0; OffX: TReal = 0; OffY: TReal = 0; SpecularColor: TColor = 0; FadeAlpha: TReal = 1; DirX: TReal = 0; DirY: TReal = -1; Narrowness: TReal = 0);
var
  X, Y: Integer;
  nX, nY, xX, xY: Integer;
  dX, dY, hX, hY: TReal;
  R2, iR2: TReal;
  D2, nD2, nAlpha: TReal;
  pBC: PBGRFColor;
  BC: TBGRFColor;
  C: TRGBFColor;
  ccDot, ccBG, ccSpecular, ccSolid: TReal;
  FR, FG, FB: TReal;
  bDraw: Boolean;
  ErrStr: String;

begin
  try
  nX := Floor(CenterX - Rad);
  xX :=  Ceil(CenterX + Rad);
  nY := Floor(CenterY - Rad);
  xY :=  Ceil(CenterY + Rad);
  except
    ErrStr := FloatToStr(CenterX) + ', ' + FloatToStr(CenterY) + '+/-' + FloatToStr(Rad);
    raise;
  end;
  if nX <  0     then nX := 0;
  if xX >= Width then xX := Width - 1;
  if nX > xX then Exit;
  if nY <  0      then nY := 0;
  if xY >= Height then xY := Height - 1;
  if nY > xY then Exit;

  hX := CenterX + OffX * Rad;
  hY := CenterY + OffY * Rad;
  R2 := (Rad + 0.5)*(Rad + 0.5);
  if dtOffset in Style then
    iR2 := 1 / Sqr(Rad * (1 + Max(Abs(OffX), Abs(OffY))) + 0.5)
  else
    iR2 := 1 / R2
  ;
  ccSolid := 0;
  ccSpecular := 0;
  for Y := nY to xY do begin
    dY := Y - CenterY;
    pBC := FindScanPixel(nX, Y);
    for X := nX to xX do begin
      dX := X - CenterX;
      D2 := dX*dX + dY*dY;
      bDraw := D2 < R2;
      if (Style * [dtDirectional, dtBidirectional] <> []) and bDraw then begin
        if dtBidirectional in Style then
          bDraw := Abs(dX * DirX + dY * DirY) > Narrowness * SqRt(D2)
        else
          bDraw := (dX * DirX + dY * DirY) > Narrowness * SqRt(D2)
        ;
      end;
      if bDraw then begin
        if dtOffset in Style then
          D2 := (Sqr(X - hX) + Sqr(Y - hY)) * iR2
        else
          D2 := D2 * iR2
        ;
        if D2 > 1 then D2 := 1;

        if dtSpecular in Style then begin
          if D2 < 1/4 then begin
            D2 := D2 * 4;
            ccSpecular := (1 - D2);
            ccDot := D2;
            ccBG := 0;
          end
          else begin
            D2 := (D2 - 1/4) * 4/3;
            ccSpecular := 0;
            ccDot := 1 - D2;
            ccBG := D2;
          end;
        end
        else begin
          ccDot := 1 - D2;
          ccBG := D2;
        end;

        if dtSolid in Style then begin
          ccSolid := ccBG;
          ccBG := 0;
        end;

        if dtFaded in Style then begin
          ccDot := ccDot * FadeAlpha;
          ccSpecular := ccSpecular * FadeAlpha;
          ccSolid := ccSolid * FadeAlpha;
          ccBG := ccBG * FadeAlpha + (1 - FadeAlpha);
        end;

        FR := 0;
        FG := 0;
        FB := 0;

        if ccBG <> 0 then begin
          BC := pBC^;
          FR := ccBG * BC.Red;
          FG := ccBG * BC.Green;
          FB := ccBG * BC.Blue;
        end;

        if ccDot <> 0 then begin
          C.Color := Color;
          FR := FR + ccDot * C.Red;
          FG := FG + ccDot * C.Green;
          FB := FB + ccDot * C.Blue;
        end;

        if ccSpecular <> 0 then begin
          C.Color := SpecularColor;
          FR := FR + ccSpecular * C.Red;
          FG := FG + ccSpecular * C.Green;
          FB := FB + ccSpecular * C.Blue;
        end;

        if ccSolid <> 0 then begin
          C.Color := SolidColor;
          FR := FR + ccSolid * C.Red;
          FG := FG + ccSolid * C.Green;
          FB := FB + ccSolid * C.Blue;
        end;

        BC.Flags := 0;
        BC.Red := Round(FR);
        BC.Green := Round(FG);
        BC.Blue := Round(FB);
        pBC^ := BC;
      end;

      Inc(pBC);
    end;
  end;
end;

procedure TScanBitmap.SubRenderPixel(X, Y: TReal; Color: TColor);
var
  iX, iY: Integer;
  dX, dY, ndX, ndY: TReal;
  CB: TRGBFColor absolute Color;
  C: TBGRFColor;
  pC: PBGRFColor;

begin
  iX := Floor(X); dX := X - iX; ndX := 1 - dX;
  iY := Floor(Y); dY := Y - iY; ndY := 1 - dY;

  pC := FindScanPixel(iX, iY);
  if pC <> nil then begin
    C := pC^;
    C.Red   := Round(C.Red   * dX * dY + CB.Red   * ndX * ndY);
    C.Green := Round(C.Green * dX * dY + CB.Green * ndX * ndY);
    C.Blue  := Round(C.Blue  * dX * dY + CB.Blue  * ndX * ndY);
    pC^ := C;
  end;
  pC := FindScanPixel(iX + 1, iY);
  if pC <> nil then begin
    C := pC^;
    C.Red   := Round(C.Red   * ndX * dY + CB.Red   * dX * ndY);
    C.Green := Round(C.Green * ndX * dY + CB.Green * dX * ndY);
    C.Blue  := Round(C.Blue  * ndX * dY + CB.Blue  * dX * ndY);
    pC^ := C;
  end;
  pC := FindScanPixel(iX, iY + 1);
  if pC <> nil then begin
    C := pC^;
    C.Red   := Round(C.Red   * dX * ndY + CB.Red   * ndX * dY);
    C.Green := Round(C.Green * dX * ndY + CB.Green * ndX * dY);
    C.Blue  := Round(C.Blue  * dX * ndY + CB.Blue  * ndX * dY);
    pC^ := C;
  end;
  pC := FindScanPixel(iX + 1, iY + 1);
  if pC <> nil then begin
    C := pC^;
    C.Red   := Round(C.Red   * ndX * ndY + CB.Red   * dX * dY);
    C.Green := Round(C.Green * ndX * ndY + CB.Green * dX * dY);
    C.Blue  := Round(C.Blue  * ndX * ndY + CB.Blue  * dX * dY);
    pC^ := C;
  end;
end;

procedure TScanBitmap.DrawDot(CenterX, CenterY, Rad: TReal; const Params: TDotParams);
begin
  DrawDot(
    CenterX, CenterY, Rad,
    Params.Color,
    Params.Style,
    Params.SolidColor,
    Params.OffX, Params.OffY,
    Params.SpecularColor,
    Params.FadeAlpha,
    Params.DirX, Params.DirY,
    Params.Narrowness
  );
end;

end.
